"""
This is the top level API for building the database required for FLAP.
The raw database file can be either a zip file of zips of csv files or a zip of gpkg file.
The delivery point address will be extracted and a SQLite database created to be used by matching.
"""


from flap.database.sql import SqlDB, SqlDBManager
from shutil import copy


def create_database(db_path, raw_db_file):
    """
    Creates the SQLite database and index the database from zip files download from OS ABP. The raw database file can be
    either a zip file of zips of csv files or a zip of gpkg file. The delivery point address will be extracted and a
    SQLite database created to be used by matching.

    Parameters
    ----------
    db_path: str
        the directory for the database
    raw_db_file
        the path to the raw zip file
    Returns
    -------
        None

    Examples
    --------
    >>> from flap import create_database
    >>> create_database(<db_path>, <raw_db_file>)
    """

    dm = SqlDBManager()
    dm.create_db(db_path)

    sql_db = SqlDB(db_path)

    copy(raw_db_file, sql_db.sub_paths['raw'])

    sql_db.build()



"""
Process of DB build:

Read raw data

Create Range expansion

Index raw data and range expansion

Expansion

"""

import os
import json
from tqdm import tqdm
import itertools
import flap


MODULE_PATH = os.path.dirname(flap.__file__)


def compile_pc1_uniques(sql_db):
    batch_gen = sql_db.sql_table_batch_by_column(
        table_name='indexed', by_columns='pc0', batch_size=int(1e5)
        )

    columns_to_compile = ["THOROUGHFARE", "DOUBLE_DEPENDENT_LOCALITY", "DEPENDENT_LOCALITY", "POST_TOWN"]

    pc1_mappings = {}

    for col in columns_to_compile:
        pc1_mappings[col] = {}

    for df in tqdm(batch_gen):

        for pc0, group in df.groupby('pc0'):

            for col in columns_to_compile:

                uniques = group[col].unique().tolist()

                try:
                    uniques.remove('')
                except ValueError:
                    pass

                if len(uniques):
                    pc1_mappings[col][pc0] = uniques

    with open(sql_db.sub_paths['pc1_mappings'], 'w') as f:
        json.dump(pc1_mappings, f, indent=4)


def compile_pc1_mappings_region(sql_db):
    path_pc1_region_mappings = os.path.join(MODULE_PATH, 'parser', 'pc1_mappings_region.json')
    os.system('cp %s %s' % (path_pc1_region_mappings, sql_db.sub_paths['pc1_mappings_region']))


def compile_global_uniques(sql_db):
    columns_to_compile = ["DOUBLE_DEPENDENT_LOCALITY", "DEPENDENT_LOCALITY", "POST_TOWN"]

    for col in columns_to_compile:
        res_sql = sql_db.sql_query("""
                                   SELECT DISTINCT %s
                                   FROM raw
                                   """ % col)
        res = list(itertools.chain(*res_sql))

        try:
            res.remove('')
        except ValueError:
            pass

        with open(sql_db.sub_paths['unique_%s' % col], 'w') as f:
            json.dump(res, f, indent=4)



"""
Consider split to different files later

"""

import pandas as pd
import numpy as np
from .cpu_count import available_cpu_count

# Span and token operations

class Span:

    def __init__(self, start, end):
        self.start, self.end = start, end
        assert self.start <= self.end, 'end < start is not allowed'

    def __contains__(self, item):

        if isinstance(item, type(self)):
            return (self.start <= item.start) and (self.end > item.end)
        elif isinstance(item, int) or isinstance(item, float) or isinstance(item, np.number):
            return (self.start <= item) and (self.end > item)
        else:
            raise TypeError('Can only compare a number or a Span')

    def __len__(self):
        return self.end - self.start

    def __str__(self):
        return 'Span: [%s, %s)' % (self.start, self.end)

    def __repr__(self):
        return 'Span(%s, %s)' % (self.start, self.end)


def tokens_to_span_classes(tokens):
    l_spans = []

    p = 0

    for token in tokens:
        l_spans.append(Span(p, p + len(token) - 1))

        p += len(token)

    return l_spans


def tokens_to_spans(tokens):
    l_spans = []

    p = 0

    for token in tokens:
        l_spans.append((p, p + len(token)))

        p += len(token)

    return l_spans


# Tokeniser

def tokenize_by_re(pattern, s):
    import re

    if isinstance(s, str):
        return [token for token in re.split(pattern, s) if len(token)]
    else:
        return ['non_str']


def tokenize_default(s):
    pattern = r'([^\w_]|\d+)'
    return tokenize_by_re(pattern, s)


# Math

def f1(prc, rec):
    return 2 * prc * rec / (prc + rec)


def f_beta(prc, rec, beta=2):
    return (1 + beta * beta) * prc * rec / (beta * beta * prc + rec)


# Utils

def parse_nodename(name):
    names = name.split(':')
    return names[0], names[1]


def span_to_lettermap(d_span, prior_l):
    for key in d_span:
        for span in d_span[key]:
            for i in range(*span):
                prior_l[i] = key

    return prior_l


def repl_positions(s, positions, repl='*'):
    sl = list(s)

    for p in positions:
        sl[p] = repl
    return ''.join(sl)


def repl_char_from_string(s, span, repl='*'):
    start, end = span

    return s[:start] + repl * (end - start) + s[end:]


def span_end(span):
    _, end = span
    return end


def dict_update_existingkeys(dict_to_update, new_dict):
    dict_to_update.update((k, new_dict[k]) for k in dict_to_update.keys())
    return dict_to_update


def span_to_string(span, s):
    start, end = span
    return s[start:end]


# UPRN formatting

def join_uprn_fields(d_uprn):
    # for key in d_uprn:
    #     if pd.isna(d_uprn[key]):
    #         d_uprn[key] = ''
    #     else:
    #         d_uprn[key] = str(d_uprn[key])

    return '\n'.join([
        ' | '.join([d_uprn['ORGANISATION_NAME'], d_uprn['DEPARTMENT_NAME'],
                    d_uprn['SUB_BUILDING_NAME'], d_uprn['BUILDING_NAME']]),
        ' | '.join([d_uprn['BUILDING_NUMBER'], d_uprn['DEPENDENT_THOROUGHFARE'], d_uprn['THOROUGHFARE']]),
        ' | '.join([d_uprn['DOUBLE_DEPENDENT_LOCALITY'], d_uprn['DEPENDENT_LOCALITY'],
                    d_uprn['POST_TOWN']]),
        ' | '.join([d_uprn['POSTCODE']])
    ])


def flatten(l):
    return [item for sublist in l for item in sublist]


def parse_uprn_fields(joined_fields):
    parsed = flatten(
              [ss.split(' | ') for ss in
               [s for s in joined_fields.split('\n')]]
                     )

    keys = [
        "ORGANISATION_NAME",
        "DEPARTMENT_NAME",
        "SUB_BUILDING_NAME",
        "BUILDING_NAME",
        "BUILDING_NUMBER",
        "DEPENDENT_THOROUGHFARE",
        "THOROUGHFARE",
        "DOUBLE_DEPENDENT_LOCALITY",
        "DEPENDENT_LOCALITY",
        "POST_TOWN",
        "POSTCODE"]

    return dict(zip(keys, parsed))


# File name


def timestamp_filename(filename):
    from pathlib import Path
    from time import ctime

    p = Path(filename)
    return "{0}-{2}{1}".format(Path.joinpath(p.parent, p.stem), p.suffix, ctime().replace(' ', '__').replace(':', '_'))



"""
Process of DB index:

Query raw db in chunks

Create Range expansion

Index raw data and range expansion

Expansion

"""

import pandas as pd
from tqdm import tqdm
import re
import itertools
import numpy as np

from flap.utils import flatten


def db_index(sql_db):

    print('Start indexing and augmentation')

    conn_temp = sql_db.get_conn_temp()

    tasks_gen = sql_db.sql_table_batch_by_column('raw', 'POSTCODE', int(1e4))

    for df_chunk in tqdm(tasks_gen):

        # Process chunks of raw data and save to 'indexed' table
        df_chunk = df_chunk.drop_duplicates()

        df_chunk['ex_label'] = '0'

        res_indexed = indexing_uprn(df_chunk)
        res_indexed.to_sql(name='indexed', con=conn_temp, if_exists='append', dtype='TEXT', index=False)

        try:
            # get and index range
            df_chunk_range = expand_bn(df_chunk)
            res_range_indexed = indexing_uprn(df_chunk_range)

            # add range
            res_both = pd.concat([res_indexed, res_range_indexed])
            res_both.reset_index(inplace=True)
        except KeyError:
            res_both = res_indexed

        # expand both and save to 'expanded'
        res_expanded = expand_number_like(res_both)
        res_expanded.to_sql(name='expanded', con=conn_temp, if_exists='append', dtype='TEXT', index=False)

    # Saving to main database

    for table_name in ['indexed', 'expanded']:

        print('Saving to %s' % table_name)

        sql_chunk_gen = pd.read_sql(sql='SELECT * FROM %s' % table_name, con=conn_temp, chunksize=int(1e5))

        conn = sql_db.get_conn()

        for chunk in tqdm(sql_chunk_gen):
            chunk.to_sql(name=table_name,
                         con=conn,
                         if_exists='append',
                         dtype='TEXT',
                         index=False
                         )

    conn.close()
    conn_temp.close()

    print('Finished database indexing')


# Below are indexing functions

NUMBER_LIKE_MASTER_REGEX = re.compile(
    r"(FLAT|UNIT|BUILDING|ROOM|BLOCK|BONDS?|FL|APARTMENT|\(F|F|-|\()? ?"
    r"((\dF\d+)|(\d+[A-Z]\d+)|([A-EG-Z])?(\d+)([A-Z])?|(?<!')(^|\b)([A-Z]|PF\d?|BF\d?|GF\d?)($|\b)(?!'))"
    r"\)?")


sbn_det_dict = {
    'level_det': ['BASEMENT', 'BOTTOM', 'GROUND', 'LOWER', 'FIRST', 'SECOND', 'THIRD',
                  'FOURTH', 'MIDDLE', 'MID', 'UPPER', 'ATTIC', 'TOP'],
    'seperator': ['APARTMENT', 'FLOOR', 'FLAT'],
    'flat_det': ['FRONT', 'REAR', 'LEFT', 'RIGHT', 'CENTRE']
}

sbn_det_tokens = []

for k, v in sbn_det_dict.items():
    sbn_det_tokens.extend(v)


sbn_parse_rule = {
    'level_det': r"%s" % '|'.join(sbn_det_dict['level_det']),
    'flat_det': r"%s" % '|'.join(sbn_det_dict['flat_det'])
}


def type_of_macro(row):

    cols = ['DEPENDENT_THOROUGHFARE', 'THOROUGHFARE', 'DOUBLE_DEPENDENT_LOCALITY', 'DEPENDENT_LOCALITY']

    return ''.join([str(int(len(row[col]) > 0)) for col in cols])


def type_of_micro(row):

    cols = ['ORGANISATION_NAME', 'DEPARTMENT_NAME', 'SUB_BUILDING_NAME',
            'BUILDING_NAME', 'BUILDING_NUMBER']

    return ''.join([str(int(len(row[col]) > 0)) for col in cols])


def parse_sbn_det(s):

    m_level = re.search(sbn_parse_rule['level_det'], s)
    m_flat = re.search(sbn_parse_rule['flat_det'], s)

    try:
        level = m_level.group(0)
    except AttributeError:
        level = ''

    try:
        flat = m_flat.group(0) if m_flat is not None else ''

    except AttributeError:
        flat = ''

    return level, flat


def all_tokens_in_all(s):
    list_of_tokens = re.split(r' ', s)
    return all([t in sbn_det_tokens for t in list_of_tokens])


def parse_number_like_uprn(row):

    p = NUMBER_LIKE_MASTER_REGEX

    split_pattern = re.compile(r"(?<=\d)(?=\D)|(?=\d)(?<=\D)")

    units = list()

    if len(row['SUB_BUILDING_NAME']):
        matches = list(re.finditer(p, row['SUB_BUILDING_NAME']))
        units.append(
            flatten([re.split(split_pattern, match.group(2))
                     if re.search(r'\dF\d', match.group(2)) is None else [match.group(2)]
                     for match in matches]))

        if all_tokens_in_all(row['SUB_BUILDING_NAME']):
            level, flat = parse_sbn_det(row['SUB_BUILDING_NAME'])
            units.append([level, flat])

    if len(row['BUILDING_NAME']):
        matches = list(re.finditer(p, row['BUILDING_NAME']))
        units.append(
            flatten([re.split(split_pattern, match.group(2))
                     if re.search(r'\dF\d', match.group(2)) is None else [match.group(2)]
                     for match in matches]))

        if all_tokens_in_all(row['SUB_BUILDING_NAME']):
            level, flat = parse_sbn_det(row['SUB_BUILDING_NAME'])
            units.append([level, flat])

    if len(row['BUILDING_NUMBER']):
        units.append([row['BUILDING_NUMBER']])

    res = flatten(units[::-1])

    if len(res) < 5:
        res += [''] * (5 - len(res))
    elif len(res) > 5:
        res = res[:5]

    return pd.Series(res)


def split_postcode_uprn(row):

    p = re.compile(r'([A-Z]{1,2})([0-9][A-Z0-9]?)(?: +)?([0-9])([A-Z])([A-Z])')

    match = re.match(p, row['POSTCODE'])

    pc0, pc1 = ''.join([match.group(i) for i in range(1, 3)]), ''.join([match.group(i) for i in range(3, 6)])

    pc_area, pc_district, pc_sector, pc_unit_0, pc_unit_1 = [match.group(i) for i in range(1, 6)]

    return pd.Series([pc0, pc1, pc_area, pc_district, pc_sector, pc_unit_0, pc_unit_1])


def count_n_tenement(df):

    df['n_tenement'] = 1

    groups = []

    for (i, j), group in df.groupby(['BUILDING_NAME', 'POSTCODE']):
        if (i != '') and (len(group) > 1):
            group['n_tenement'] = len(group)

        groups.append(group)

    df = pd.concat(groups)

    groups = []

    for (i, j), group in df.groupby(['BUILDING_NUMBER', 'POSTCODE']):
        if (i != '') and (len(group) > 1):
            group['n_tenement'] = len(group)
        groups.append(group)

    df = pd.concat(groups)

    return df


def indexing_uprn(group):

    group = count_n_tenement(group)

    group['type_of_micro'] = group.apply(type_of_micro, axis=1)
    group['type_of_macro'] = group.apply(type_of_macro, axis=1)

    group[['number_like_%s' % i for i in range(5)]] = group.apply(parse_number_like_uprn, axis=1)
    group[['pc0', 'pc1', 'pc_area', 'pc_district', 'pc_sector', 'pc_unit_0', 'pc_unit_1']] = \
        group.apply(split_postcode_uprn, axis=1)

    return group


# Code below is for range expansion


def parse_range(text):
    p = re.compile(r'^([A-Z ]+)?(\d+)([A-Z])?-(\d+)([A-Z])?([A-Z ]+)?$')

    match = re.match(p, text)

    if match is None:
        return None

    else:
        parsing_seq = ['prefix', 'num_0', 'char_0', 'num_1', 'char_1', 'suffix']
        parsed = {name: group for name, group in zip(parsing_seq, match.groups())}
        return parsed


def range_letter_upper(letter_1, letter_2):
    l2n = {chr(x): x for x in range(65, 91)}
    return [chr(x) for x in range(l2n[letter_1], l2n[letter_2] + 1)]


def expand_str(ll):
    ll = [l for l in ll if len(l)]
    return [''.join(i) for i in itertools.product(*ll)]


def bn_range_to_list(row):

    text = row['BUILDING_NAME']

    parsed = parse_range(text)

    if parsed is None:
        return []

    else:
        pre_expand = {
            'prefix': [] if parsed['prefix'] is None else [parsed['prefix']],
            'range': [],
            'suffix': [] if parsed['suffix'] is None else [parsed['suffix']]
        }

        interval = 1 if ((int(parsed['num_0']) + int(parsed['num_1'])) % 2) or \
                        (parsed['prefix'] is not None) or \
                        (parsed['suffix'] is not None) else 2

        if (parsed['char_0'] is None) and (parsed['char_1'] is None):

            pre_expand['range'] = [str(i)
                                   for i in range(int(parsed['num_0']), int(parsed['num_1']) + interval, interval)]

        elif parsed['char_0'] is None:

            if parsed['num_0'] == parsed['num_1']:
                pre_expand['range'] = [parsed['num_0'], ''.join([parsed['num_1'], parsed['char_1']])]
            else:
                pre_expand['range'] = [
                    str(i)
                    for i in range(int(parsed['num_0']), int(parsed['num_1']), interval)
                    ] + [''.join([parsed['num_1'], parsed['char_1']])]

        elif (parsed['char_0'] is not None) and (parsed['char_1'] is not None):

            if parsed['num_0'] == parsed['num_1']:

                pre_expand['range'] = [
                    ''.join([parsed['num_0'], letter])
                    for letter in range_letter_upper(parsed['char_0'], parsed['char_1'])
                ]

        expand_mappings = []

        if len([pre_expand['range']]):

            for s in expand_str(list(pre_expand.values())):
                if re.search(r'[A-Z]', s) is not None:
                    expand_mappings.append({'BUILDING_NAME': s})
                else:
                    expand_mappings.append({'BUILDING_NAME': '', 'BUILDING_NUMBER': s})

        if isinstance(row, dict):
            ref_row_d = row
        else:
            ref_row_d = row.to_dict()

        rows = []

        for mapping in expand_mappings:
            rows.append(dict(ref_row_d,
                             ex_label='-'.join([ref_row_d['ex_label'], '1rng']),
                             **mapping))

    return rows


def expand_bn(group):

    expand_collections = []

    for _, row in group.iterrows():
        ex = bn_range_to_list(row)
        expand_collections.extend(ex)

    return pd.DataFrame.from_records(expand_collections)


# Code below is for expansion

# # Ranking and Mapping functions


def parse_multiplier(x):
    m = re.search(r'(\d+)F(\d+)', x)
    level = int(m.group(1))
    flat = int(m.group(2))
    return level, flat


def multiplier_to_numeric(level, flat, max_flat=20):
    return (level - 1) * max_flat + flat


def multiplier_to_number(level, flat, max_flat):
    return (level - 1) * max_flat + flat


def rank_multiplier(x):
    parsed = [parse_multiplier(s) for s in x]
    numbered = [multiplier_to_numeric(level, flat) for level, flat in parsed]
    rank = list(np.argsort(np.argsort(numbered)) + 1)
    return rank


def rank_lcr(x):
    x = x.to_list()
    if 'C' not in x:
        mapping = {'L': '1', 'R': '2'}
        return [mapping[s] for s in x]
    elif len(x) == 1:
        return ['3']
    else:
        mapping = {'L': '1', 'C': '2', 'R': '3'}
        return [mapping[s] for s in x]


def rank_a2z(x):

    if 'I' not in x:
        letters = range_letter_upper('A', 'H') + range_letter_upper('J', 'Z')
    else:
        letters = range_letter_upper('A', 'Z')
    mapping = {s: str(i + 1) for i, s in enumerate(letters)}

    return [s if s not in mapping else mapping[s] for s in x]


def rank_g(x):

    mapping = {'G': 0}

    return [s if s not in mapping else mapping[s] for s in x]


def abbr_flat_det(x):
    if 'C' not in x:
        mapping = {
            'LEFT': 'L;1',
            'RIGHT': 'R;2'
        }
    else:
        mapping = {
            'LEFT': 'L;1',
            'RIGHT': 'R;3',
            'CENTRE': 'C;2'
        }

    return [s if s not in mapping else mapping[s] for s in x]


def abbr_level_det(x):
    mapping = {
        'BASEMENT': 'BF',
        'BOTTOM': 'BF',
        'GROUND': 'G;0',
        'FIRST': '1',
        'SECOND': '2',
        'THIRD': '3',
        'FOURTH': '4',
        'MIDDLE': 'M',
        'MID': 'M',
        'ATTIC': '<max>',
        'TOP': '<max>',
    }

    return [s if s not in mapping else mapping[s] for s in x]


# # Calculation for local max


def local_max_det(series):
    all_numbers = []

    for s in series:

        ss = s.split(';')
        for sss in ss:
            if sss.isnumeric():
                all_numbers.append(int(sss))

    if len(all_numbers):
        return max(all_numbers)
    else:
        return 0


sbn_det_dict = {
    'level_det': ['BASEMENT', 'BOTTOM', 'GROUND', 'LOWER', 'FIRST', 'SECOND', 'THIRD',
                  'FOURTH', 'MIDDLE', 'MID', 'UPPER', 'ATTIC', 'TOP'],
    'seperator': ['APARTMENT', 'FLOOR', 'FLAT'],
    'flat_det': ['FRONT', 'REAR', 'LEFT', 'RIGHT', 'CENTRE']
}


expansion_rules = {
    'multiplier': {
        'inclusion': r'^\d+F\d+$',
        'exclusion': r'^\d+$',
        'mapping': rank_multiplier
    },
    'level_det': {
        'inclusion': r'%s' % '|'.join(sbn_det_dict['level_det']),
        'exclusion': r'99999',
        'mapping': abbr_level_det
    },
    'flat_det': {
        'inclusion': r'%s' % '|'.join(sbn_det_dict['flat_det']),
        'exclusion': r'(^d+$)|(^[A-Z]$)',
        'mapping': abbr_flat_det
    },
    'lcr': {
        'inclusion': r'^(L|C|R)$',
        'exclusion': r'(^\d+$)|(^[ABD-KM-QT-Z]$)',
        'mapping': rank_lcr
    },
    'a_to_z': {
        'inclusion': r'^[A-Z]$',
        'exclusion': r'^\d+$',
        'mapping': rank_a2z
    },
    'G': {
        'inclusion': r'^G$',
        'exclusion': r'^[A-FH-Z]$',
        'mapping': rank_g
    }

}


def expand_number_like(group):

    levels = ['number_like_%s' % s for s in range(5)]
    expansion_collection = {k: [] for k in levels}

    expanded = False

    for i in range(1, 5):

        # Start of Level loop

        group_by = levels[:i]

        if len(group_by) == 1:
            group_by = group_by[0]

        for indices, gg in group.groupby(group_by):

            # Start of GroupBy/GG loop

            gg_alt = gg.copy()

            for k, v in expansion_rules.items():

                # Start of Rule Loop

                incl_match = [re.search(r'%s' % v['inclusion'], s) for s in gg[levels[i]]]
                excl_match = [re.search(r'%s' % v['exclusion'], s) for s in gg[levels[i]]]

                incl = [m is not None for m in incl_match]
                excl = [m is not None for m in excl_match]

                if any(incl) and (not any(excl)):

                    mapped = v['mapping'](gg_alt.loc[incl, levels[i]])

                    # print(mapped)

                    gg_alt.loc[incl, levels[i]] = [
                        ';'.join([x, str(y)]) for x, y in zip(gg_alt.loc[incl, levels[i]], mapped)]

                    expanded = True

                # End of Rule Loop

            if any(['<max>' in s for s in gg_alt[levels[i]]]):

                max_det = local_max_det(gg_alt[levels[i]])

                gg_alt[levels[i]] = gg_alt[levels[i]].apply(lambda x: x.replace('<max>', str(max_det + 1)))

            expansion_collection[levels[i]].append(gg_alt[[levels[i]]])

            # End of GroupBy/GG loop

        expansion_collection[levels[i]] = pd.concat(expansion_collection[levels[i]])

        # End of level loop

    if expanded:

        expansion_collection['number_like_0'] = group[['number_like_0']]

        group.loc[:, levels] = pd.concat([df for df in expansion_collection.values()], axis=1)

        return pd.DataFrame.from_records(flatten([expand_number_like_row(row) for _, row in group.iterrows()]))

    else:
        return pd.DataFrame([])


def expand_list(ll):
    return [list(item) for item in itertools.product(*ll)]


def expand_dict(d):
    ll = expand_list(d.values())
    dl = []

    for l in ll:
        dl.append({k: v for k, v in zip(d.keys(), l)})

    return dl


def expand_number_like_row(row):

    levels = ['number_like_%s' % s for s in range(5)]

    row_d = row.to_dict()
    to_expand = {col: list(set(row_d[col].split(';')[1:])) for col in levels if ';' in row_d[col]}

    l_ex = expand_dict(to_expand)

    d_ex = []

    if len(l_ex[0]):
        for d in l_ex:
            row_d.update(d)
            row_d['ex_label'] = row_d['ex_label'] + '_2_nl'
            d_ex.append(row_d)

    return d_ex



import pandas as pd
import os

from tqdm import tqdm

from flap.api.score import score
from flap.api.match import match


def score_and_match(input_csv, db_path, final_output=None,
                    max_workers=None, in_memory_db=False, classifier_model_path=None,
                    input_address_col='input_address', uprn_col='uprn',
                    score_output=None, score_output_raw=None, score_batch_size=100000,
                    match_output=None, match_output_raw=None, match_batch_size=10000,
                    match_in_progress_log_path=None, match_max_log_interval=4800,
                    max_beam_width=200, score_threshold=0.3):

    """
    This is the workflow to score existing matches and match the remaining ones and ones with scores below 0.5

    Parameters
    ----------
    input_csv : str, or pandas.DataFrame
        Path to the csv file. The file needs to have two fields in the header ['input_id', 'input_address']
    db_path : str
        Path to the database built. See `flap.create_db()`
    final_output: str, default None
        Path for saving the final output csv file, containing ['input_id', 'input_address', 'uprn', 'score'].
        If None, results are not saved
    max_workers : int, default None
        Number of processes. If None, the max cpu available is determined by `flap.utils.cpu_count.available_cpu_count()`
    in_memory_db : bool, default False
        If in-memory SQLite database is used. If True, a temp database is created in shared memory cache from pre-built
        csv files
    classifier_model_path : str, default None
        The path to the pretrained sklearn classifier model.
        If None, the model is loaded from 'flap.__file__/model/*.clf'
    input_address_col: str, default 'input_address'
        The column name for input addresses for score function
    uprn_col: str, default 'uprn'
        The column name for input UPRN for score function
    score_output
    score_output_raw
    score_batch_size
    match_output
    match_output_raw
    match_batch_size
    match_in_progress_log_path
    match_max_log_interval
    max_beam_width: int, default 200
        The max number of rows to be considered from UPRN database
    score_threshold: float, default 0.3
        The min score for early stop of matching

    Returns
    -------
    pandas.DataFrame
        Final results
    """

    if not os.path.exists(score_output):

        scored = score(input_csv, db_path, output_file_path=score_output, raw_output_path=score_output_raw,
                       batch_size=score_batch_size, max_workers=max_workers, in_memory_db=in_memory_db,
                       classifier_model_path=classifier_model_path,
                       input_address_col=input_address_col, uprn_col=uprn_col)

    else:
        scored = pd.read_csv(score_output, index_col=0, dtype='object')

    scored['flap_eval_score'] = scored['flap_eval_score'].astype(float)

    scored_to_flap = scored[(scored.flap_eval_score < 0.5) | scored.flap_eval_score.isna()].copy()

    print(len(scored_to_flap))

    scored_skipped = scored[scored.flap_eval_score >= 0.5].copy()
    scored_skipped['flap_match_score'] = scored_skipped['flap_eval_score']
    scored_skipped['flap_uprn'] = scored_skipped['uprn']

    matched = match(input_csv=scored_to_flap, db_path=db_path, output_file_path=match_output,
                    raw_output_path=match_output_raw,
                    in_progress_log_path=match_in_progress_log_path, max_log_interval=match_max_log_interval,
                    batch_size=match_batch_size, max_workers=max_workers, in_memory_db=in_memory_db,
                    classifier_model_path=classifier_model_path,
                    max_beam_width=max_beam_width, score_threshold=score_threshold
                    )

    final = pd.concat([scored_skipped, matched])

    if final_output is not None:
        final.to_csv(final_output)

    return final



"""
DB_read module extract delivery point address from Address Premium Product

Currently, it accepts two types: zip of zips of csvs (.zip) or a geopackage (.gpkg)

A 'raw' table will be created in the Sqlite database
"""

from zipfile import ZipFile
import csv
import io
import pandas as pd
import geopandas as gpd
import os
from tqdm import tqdm


def numeric_to_object(x):
    if pd.isna(x):
        return ''

    elif isinstance(x, int):
        return str(x)

    elif isinstance(x, float):
        return str(int(x))

    else:
        return x


def convert_df_to_object(df):
    for col in df:

        if pd.api.types.is_object_dtype(df[col]):
            df.loc[:, col] = df[col].fillna('')

        elif pd.api.types.is_numeric_dtype(df[col]):
            df.loc[:, col] = df[col].apply(lambda x: numeric_to_object(x))

        else:
            df.loc[:, col] = df[col]

    return df


def db_import(sql_db):
    
    sql_db.drop_table('raw')

    conn = sql_db.get_conn()

    files_in_raw = [os.path.join(sql_db.sub_paths['raw'], file)
                    for file in os.listdir(sql_db.sub_paths['raw'])]

    assert len(files_in_raw) == 1, 'There should be only one raw DB file in the raw folder'

    raw_file = files_in_raw[0]

    assert ('.csv' in raw_file) or ('.zip' in raw_file), 'Raw DB file should be either a .csv file or a .zip file'

    if '.csv' in raw_file:

        headers = pd.read_csv(raw_file, dtype='object', nrows=1, header=None)

        if headers.iloc[0, 0] == '28':
            names = list(sql_db.get_db_config().keys())[:len(headers.columns)]

            print('Please check the fields: \n')
            sample_fields = {k: v for k, v in zip(names, headers.iloc[0, :].tolist())}
            print(sample_fields)

            raw = pd.read_csv(raw_file, names=names,
                              chunksize=int(1e5), dtype='object')
            
        else:
            headers.fillna('', inplace=True)
            index_col = 0 if headers.iloc[0, 0] == '' else None
            raw = pd.read_csv(raw_file, chunksize=int(1e5), dtype='object', index_col=index_col)

        for df_chunk in tqdm(raw):
            df_chunk.fillna('', inplace=True)
            df_chunk.to_sql(name='raw', con=conn, if_exists='append', dtype='TEXT', index=False)

        return

    zip_path = [os.path.join(sql_db.sub_paths['raw'], file)
                for file in os.listdir(sql_db.sub_paths['raw']) if '.zip' in file][0]

    with ZipFile(zip_path, 'r') as z:

        zz_files = [zz_file for zz_file in z.namelist()]

        if any('.gpkg' in zz_file for zz_file in zz_files):

            print('.gpkg database found. Starting to read db file')

            zz_data = io.BytesIO(z.read(zz_files[0]))

            gpkg_db = gpd.read_file(zz_data, layer='delivery_point_address')

            print('DB reading complete, start saving to SQL')

            gpkg_db = convert_df_to_object(gpkg_db)

            cols_upper = [col.upper() for col in gpkg_db.columns]

            gpkg_db.columns = cols_upper

            gpkg_db.to_sql(name='raw', con=conn, if_exists='append', dtype='TEXT', index=False)

            print('Raw table building finished')

        elif any('.zip' in zz_file for zz_file in zz_files):

            print('.zip database found. Starting to read db file')

            chunk_size = int(1e5)
            rows = []
            data_index = '28'

            sql_config = sql_db.get_db_config()

            zz_files = [zz_file for zz_file in z.namelist() if '.zip' in zz_file]

            for zz_file in tqdm(zz_files):
                zz_data = io.BytesIO(z.read(zz_file))

                with ZipFile(zz_data, 'r') as zz:
                    csv_files = [csv_file for csv_file in zz.namelist() if '.csv' in csv_file]

                    for csv_file in csv_files:

                        with zz.open(csv_file, 'r') as f:

                            lines = [line.decode() for line in f.readlines()]

                            reader = csv.reader(lines)

                            for row in reader:
                                if row[0] == data_index:
                                    rows.append(row)

                        while len(rows) > chunk_size:
                            df = pd.DataFrame.from_records(rows, columns=list(sql_config.keys()))
                            df.to_sql(name='raw', con=conn, if_exists='append', dtype=sql_config, index=False)
                            rows = rows[chunk_size:]

            df = pd.DataFrame.from_records(rows, columns=list(sql_config.keys()))
            df.to_sql(name='raw', con=conn, if_exists='append', dtype=sql_config, index=False)
            print('Raw table building finished')

        elif any('.csv' in zz_file for zz_file in zz_files):

            print('.csv database found. Starting to read db file')

            csv_files = [zz_file for zz_file in zz_files if '.csv' in zz_file]

            sql_config = sql_db.get_db_config()

            for file in csv_files:
                with z.open(file) as f_csv:

                    print('Processing %s' % file)

                    df = pd.read_csv(f_csv, header=None, dtype='object', names=list(sql_config.keys()))

                    df.to_sql(name='raw', con=conn, if_exists='append', dtype='TEXT', index=False)

            print('Raw table building finished')

        else:
            raise 'Format of the raw file is not recognised.'

    conn.close()



"""
The flap.database.sql module contains classes handling creating, connecting and querying SQL database.

SQL database is used for FLAP for its ability to handle large database and to do fast querying via reverse indexing.
"""

import pandas as pd
import json
import os
import sqlite3
import shutil
from warnings import warn
from tqdm import tqdm

import flap

from flap.database.db_import import db_import
from flap.database.db_build_vocabulary import compile_pc1_uniques, compile_pc1_mappings_region, compile_global_uniques
from flap.database.db_index import db_index


MODULE_PATH = os.path.dirname(flap.__file__)


class SqlDBManager:

    def __init__(self, project_name=None):
        self.global_db_path = os.path.join(MODULE_PATH, 'db')
        self.project_name = project_name
        if project_name is not None:
            self.project_db_path = os.path.join(os.getcwd(), project_name, 'db')

    def list_global_db(self):
        db_names = os.listdir(self.global_db_path)
        return db_names

    def list_project_db(self):
        db_names = os.listdir(self.project_db_path)
        return db_names

    @staticmethod
    def create_db(path):
        new_db_path = path

        db_name = os.path.basename(path)

        if os.path.exists(new_db_path):
            print('DB:%s exists, the existing database is used.' % db_name)
        else:
            os.makedirs(new_db_path, exist_ok=True)
            print('New db created at: %s' % new_db_path)
            os.mkdir(os.path.join(new_db_path, 'raw'))
            os.mkdir(os.path.join(new_db_path, 'sql'))
            os.mkdir(os.path.join(new_db_path, 'db_config'))
            os.mkdir(os.path.join(new_db_path, 'csv'))

    def create_global_db(self, db_name):
        new_db_path = os.path.join(self.global_db_path, db_name)

        if os.path.exists(new_db_path):
            warn('DB:%s exists, new database cannot be created.' % db_name)
        else:
            os.makedirs(new_db_path, exist_ok=True)
            print('New db created at: %s' % new_db_path)
            os.mkdir(os.path.join(new_db_path, 'raw'))
            os.mkdir(os.path.join(new_db_path, 'sql'))
            os.mkdir(os.path.join(new_db_path, 'db_config'))
            os.mkdir(os.path.join(new_db_path, 'csv'))

    def create_project_db(self, db_name):
        new_db_path = os.path.join(self.project_db_path, db_name)

        if os.path.exists(new_db_path):
            warn('DB:%s exists, new database cannot be created.' % db_name)
        else:
            os.makedirs(new_db_path, exist_ok=True)
            print('New db created at: %s' % new_db_path)
            os.mkdir(os.path.join(new_db_path, 'raw'))
            os.mkdir(os.path.join(new_db_path, 'sql'))
            os.mkdir(os.path.join(new_db_path, 'db_config'))
            os.mkdir(os.path.join(new_db_path, 'csv'))

    def get_db_path(self, db_name, project_level=False):
        if not project_level:
            if db_name in self.list_global_db():
                return os.path.join(self.global_db_path, db_name)
            else:
                warn('DB: %s does not exist!' % db_name)
        else:
            if self.project_name is None:
                warn('Project not specified')
            else:
                if db_name in self.list_project_db():
                    return os.path.join(self.project_db_path, db_name)
                else:
                    warn('DB: %s does not exist!' % db_name)

    def get_db(self, db_name, project_level=False):

        return SqlDB(self.get_db_path(db_name, project_level))


class SqlDB:
    """
    SqlDB class has the methods for creating, indexing, querying of SqlDB
    """
    def __init__(self, path_to_db):
        """
        Parameters
        ----------
        path_to_db : str,
            The path for creating the SQL database
        """
        self.db_path = path_to_db
        self.db_name = os.path.basename(self.db_path)
        self.sub_paths = {
            'raw': os.path.join(self.db_path, 'raw'),
            'db_config': os.path.join(self.db_path, 'db_config'),
            'sql_db': os.path.join(self.db_path, 'sql', 'db.sqlite3'),
            'sql_db_temp': os.path.join(self.db_path, 'sql', 'db_temp.sqlite3'),
            'csv': os.path.join(self.db_path, 'csv'),
            'vocabulary': os.path.join(self.db_path, 'vocabulary'),
            'pc1_mappings': os.path.join(self.db_path, 'vocabulary', 'pc1_mappings.json'),
            'pc1_mappings_region': os.path.join(self.db_path, 'vocabulary', 'pc1_mappings_region.json'),
            'unique_DOUBLE_DEPENDENT_LOCALITY': os.path.join(self.db_path, 'vocabulary',
                                                             'unique_DOUBLE_DEPENDENT_LOCALITY.json'),
            'unique_DEPENDENT_LOCALITY': os.path.join(self.db_path, 'vocabulary', 'unique_DEPENDENT_LOCALITY.json'),
            'unique_POST_TOWN': os.path.join(self.db_path, 'vocabulary', 'unique_POST_TOWN.json'),
            'thoroughfare_patterns': os.path.join(self.db_path, 'vocabulary', 'thoroughfare_patterns.json')
        }

        self.__db_status = None

    def __str__(self):
        self_str = f"""
        Database name: {self.db_name}
        Database path: {self.db_path}
        """
        return self_str

    def get_table_names(self):
        """
        Get names of all tables in the database

        Returns
        -------
        list
            Names of all tables in the database
        """
        conn = self.get_conn()
        cur = conn.cursor()

        try:
            cur.execute("""SELECT name FROM sqlite_schema""")
            res = cur.fetchall()

            if len(res):
                table_names = [list(name)[0] for name in res]
            else:
                table_names = []

        except sqlite3.OperationalError:
            table_names = []

        cur.close()
        conn.close()

        return table_names

    def get_columns_of_table(self, table_name):
        """
        Method to get columns of a table in the database
        Parameters
        ----------
        table_name : str
            The name of the table to be queried
        Returns
        -------
        list
            The column names of the table
        """
        table_name = (table_name,)

        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute("""
        SELECT sql FROM sqlite_master WHERE tbl_name = ? AND type = 'table'
        """, table_name)
        res = cur.fetchall()

        try:
            s = res[0][0]
            lines = s.split('\n')
            column_names = [line.split('"')[1] for line in lines[1:-1]]
        except TypeError:
            column_names = []
        except IndexError:
            column_names = []

        cur.close()
        conn.close()

        return column_names

    @property
    def db_status(self):
        """
        Print the status of the database
        Returns
        -------
        None
        """
        db_status = {
            'db_name': self.db_name,
            'db_path': self.db_path,
            'valid': os.path.isdir(self.db_path),
            'raw_added': len(os.listdir(self.sub_paths['raw'])) > 0
        }

        tables = self.get_table_names()

        db_status['table_raw_built'] = 'raw' in tables
        db_status['table_indexed_built'] = 'indexed' in tables

        self.__db_status = db_status

        return self.__db_status

    def drop_table(self, table):
        """
        Drop a table from the database
        Parameters
        ----------
        table : str
            Name of the table to be dropped
        Returns
        -------
        None
        """
        conn = self.get_conn()
        cur = conn.cursor()

        try:
            cur.execute(f"""DROP TABLE {table}""")

        except sqlite3.OperationalError:
            print('Table not exist')

        conn.commit()
        cur.close()
        conn.close()

    def get_db_config(self):
        """
        Get config of the database.
        Config can be set via `.json` file in `./db_config/*.json` otherwise default will be returned
        Returns
        -------
        """
        sql_config_path = [os.path.join(self.sub_paths['db_config'], file)
                           for file in os.listdir(self.sub_paths['db_config']) if '.json' in file]

        if len(sql_config_path):

            sql_config_file = sql_config_path[0]

            with open(sql_config_file, 'r') as f:
                sql_config = json.load(f)

        else:

            sql_config = {
                "RECORD_IDENTIFIER": "INTEGER",
                "CHANGE_TYPE": "TEXT",
                "PRO_ORDER": "TEXT",
                "UPRN": "TEXT",
                "UDPRN": "TEXT",
                "ORGANISATION_NAME": "TEXT",
                "DEPARTMENT_NAME": "TEXT",
                "SUB_BUILDING_NAME": "TEXT",
                "BUILDING_NAME": "TEXT",
                "BUILDING_NUMBER": "TEXT",
                "DEPENDENT_THOROUGHFARE": "TEXT",
                "THOROUGHFARE": "TEXT",
                "DOUBLE_DEPENDENT_LOCALITY": "TEXT",
                "DEPENDENT_LOCALITY": "TEXT",
                "POST_TOWN": "TEXT",
                "POSTCODE": "TEXT",
                "POSTCODE_TYPE": "TEXT",
                "DELIVERY_POINT_SUFFIX": "TEXT",
                "WELSH_DEPENDENT_THOROUGHFARE": "TEXT",
                "WELSH_THOROUGHFARE": "TEXT",
                "WELSH_DOUBLE_DEPENDENT_LOCALITY": "TEXT",
                "WELSH_DEPENDENT_LOCALITY": "TEXT",
                "WELSH_POST_TOWN": "TEXT",
                "PO_BOX_NUMBER": "TEXT",
                "PROCESS_DATE": "TEXT",
                "START_DATE": "TEXT",
                "END_DATE": "TEXT",
                "LAST_UPDATE_DATE": "TEXT",
                "ENTRY_DATE": "TEXT"
            }

            sql_config_file = os.path.join(self.sub_paths['db_config'], 'db_config.json')

            with open(sql_config_file, 'w') as f:
                json.dump(sql_config, f, indent=4)

            print('db_config created')
        return sql_config

    def setup_database(self, if_exists='skip'):
        """
        The wrapper for the whole setup process
        Parameters
        ----------
        if_exists : 'skip' or 'replace'
            if 'skip' raw and indexing will be skipped if exists
        Returns
        -------
        None
        """
        self.build_raw(if_exists=if_exists)
        self.indexing_db(if_exists=if_exists)
        self.build_vocabulary()
        self.build_csv()
        print('Database Setup complete!')

    def build(self, if_exists='skip'):
        """
        Synonym for `setup_database` method
        Parameters
        ----------
        if_exists : 'skip' or 'replace'
            if 'skip' raw and indexing will be skipped if exists
        Returns
        -------
        None
        """
        self.build_raw(if_exists=if_exists)
        self.indexing_db(if_exists=if_exists)
        self.build_vocabulary()
        self.build_csv()
        print('Database Setup complete!')
        
    def build_raw(self, if_exists='skip'):
        """
        Import data from the raw file and build the raw table using the function `flap.database.db_import.db_import`
        Parameters
        ----------
        if_exists : 'skip' or 'replace'
            if 'skip' raw import will be skipped if exists
        Returns
        -------
        None
        """
        db_status = self.db_status

        print(db_status)

        if not db_status['raw_added']:
            raise FileNotFoundError('Add raw database files to [db_path]/raw first and then run build')

        if if_exists == 'skip':

            if db_status['table_raw_built']:
                print('Table raw exists, build skipped. Use if_exists="replace", if you want to create a new db')
            else:
                db_import(self)

        elif if_exists == 'replace':

            self.drop_table('raw')
            db_import(self)

    def indexing_db(self, if_exists='skip'):
        """
        Index the raw table using the function `flap.database.db_index.db_index`
        Parameters
        ----------
        if_exists : 'skip' or 'replace'
            if 'skip' indexing will be skipped if `indexed` table exists
        Returns
        -------
        None
        """
        db_status = self.db_status

        if if_exists == 'skip':

            if db_status['table_indexed_built']:
                print('TABLE indexed exists, build skipped. Use if_exists="replace", if you want to create a new db')
            else:
                db_index(self)
                self.delete_temp()

        elif if_exists == 'replace':

            self.drop_table('indexed')
            db_index(self)
            self.delete_temp()

    def build_vocabulary(self):
        """
        Build vocabularies for address parsing. See `flap.parser.rule_parser_fast.RuleParserFast`
        Returns
        -------
        None
        """
        print('Start building vocabularies')
        if not os.path.exists(self.sub_paths['vocabulary']):
            os.mkdir(self.sub_paths['vocabulary'])

        compile_pc1_uniques(self)
        compile_pc1_mappings_region(self)
        compile_global_uniques(self)

        path_thoroughfare_patterns = os.path.join(MODULE_PATH, 'parser', 'thoroughfare_patterns.json')
        shutil.copy(path_thoroughfare_patterns, self.sub_paths['thoroughfare_patterns'])

        print('Finished building vocabularies')

    def build_csv(self):
        """
        Export `indexed` and `expanded` table to csv files for the use in `flap.database.sql_in_memory.SqlDBInMemory`
        Returns
        -------
        None
        """
        for table_name in ['indexed', 'expanded']:

            print(f'Start saving {table_name} to csv for in-memory DB')

            header = True

            cols = self.get_columns_of_table(table_name)

            sql = f'select * from {table_name}'

            chunk_gen = pd.read_sql(sql, con=self.get_conn(), columns=cols, chunksize=int(1e5))

            for chunk in tqdm(chunk_gen):

                chunk.to_csv(os.path.join(self.sub_paths['csv'], f'{table_name}.csv.gz'),
                             header=header, mode='a', compression='gzip')
                header = False

    def get_conn(self):
        """
        Open a connection to the SQL database
        Returns
        -------
        sqlite3.conn
        """
        return sqlite3.connect(self.sub_paths['sql_db'], timeout=10)

    def get_conn_temp(self):
        """
        Open a connection to the temporary SQL database
        Returns
        -------
        sqlite3.conn
        """
        return sqlite3.connect(self.sub_paths['sql_db_temp'], timeout=10)

    def delete_temp(self):
        """
        Delete the temporary SQL database
        Returns
        -------
        None
        """
        try:
            os.remove(self.sub_paths['sql_db_temp'])
        except FileNotFoundError:
            pass

    def sql_query(self, query):
        """
        Query the SQL database
        Parameters
        ----------
        query : str
            The query string containing SQL script
        Returns
        -------
        list
            Query results
        """
        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute(query)
        res = cur.fetchall()
        cur.close()
        conn.close()

        return res

    def sql_query_by_column_values(self, table_name, column, value_list):
        """
        Query the table in the format `f"select * from {table_name} where {column} IN ({in_clause})"`
        Parameters
        ----------
        table_name : str
        column : str
        value_list : str

        Returns
        -------
        list
            Query results
        """
        in_clause = ', '.join(value_list)
        res = self.sql_query(f"""select * from {table_name} where {column} IN ({in_clause})""")
        return res

    def sql_query_batch(self, query, batch_size=int(1e5)):
        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute(query)

        for batch in iter(lambda: cur.fetchmany(batch_size), []):
            yield batch

        conn.close()

    def sql_table_batch_by_column(self, table_name, by_columns, batch_size, match_func=None):

        if isinstance(by_columns, str):
            by_columns = [by_columns]

        table_columns = self.get_columns_of_table(table_name)

        column_indices = [table_columns.index(col) for col in by_columns]

        query = """
            SELECT *
            FROM %s
            ORDER BY %s
        """ % (table_name, ','.join(by_columns))

        con = self.get_conn()
        cur = con.cursor()
        cur.execute(query)

        res = cur.fetchmany(batch_size)
        next_res = cur.fetchone()

        while len(res):

            try:

                if match_func is None:
                    while all([next_res[i] == res[-1][i] for i in column_indices]):
                        res.append(next_res)
                        next_res = cur.fetchone()
                else:
                    while match_func(next_res, res):
                        res.append(next_res)
                        next_res = cur.fetchone()

            except TypeError:
                pass

            res_df = pd.DataFrame.from_records(res)
            res_df.columns = table_columns

            yield res_df

            if next_res is None:
                break
            else:
                res = [next_res]
                res.extend(cur.fetchmany(batch_size))
                next_res = cur.fetchone()

        con.close()

    def attach_uprn_fields_to_df(self, df, uprn_col='uprn'):

        uprn_list = [s for s in df[uprn_col].to_list() if isinstance(s, str)]

        res = self.sql_query_by_column_values(table_name='indexed', column='UPRN', value_list=uprn_list)
        columns = self.get_columns_of_table(table_name='indexed')
        res_df = pd.DataFrame.from_records(res, columns=columns)

        df = pd.merge(left=df, right=res_df, left_on=uprn_col, right_on='UPRN', how='left', suffixes=(None, '_db'))

        return df

    def create_index(self, table_name, columns):

        index_name = '__'.join(columns)

        sql = "CREATE INDEX IF NOT EXISTS %s ON %s(%s)" % (
            index_name,
            table_name,
            ', '.join(columns)
        )

        with self.get_conn() as conn:
            cur = conn.cursor()
            cur.execute(sql)



import sqlite3
import pandas as pd


class SqlDBInMemory:

    def __init__(self):
        self.columns_dict = {}
        self.update_columns_dict()

    @staticmethod
    def get_conn():
        return sqlite3.connect('file:cachedb?mode=memory&cache=shared', timeout=10)

    def update_columns_dict(self):

        conn = self.get_conn()

        cur = conn.cursor()
        cur.execute("""SELECT name FROM sqlite_master WHERE type='table'""")

        tables = [i[0] for i in cur.fetchall()]

        cur.close()

        for table_name in tables:

            cur = conn.cursor()

            cur.execute("""SELECT sql FROM sqlite_master WHERE tbl_name = ? AND type = 'table'""", (table_name,))
            res = cur.fetchall()

            cur.close()

            try:
                s = res[0][0]
                lines = s.split('\n')
                column_names = [line.split('"')[1] for line in lines[1:-1]]

            except TypeError:
                column_names = []
            except IndexError:
                column_names = []

            self.columns_dict[table_name] = column_names

    def load_csv(self, path, table_name, chunksize=10000):

        self.update_columns_dict()

        if table_name in self.columns_dict:
            print('Use Existing DB in Cache')
            return

        print('Start loading DB to memory')

        db_chunks = pd.read_csv(path, chunksize=chunksize, dtype='object', index_col=0)

        i = 0

        while 1:
            try:
                df = next(db_chunks)
                i += 1
                df.fillna('', inplace=True)
                df.to_sql(name=table_name, con=self.get_conn(), if_exists='append', dtype='TEXT', index=False)
                print('Chunk %s loaded' % i, end='\r')

                self.columns_dict[table_name] = df.columns.to_list()

            except StopIteration:
                break

        print('DB Loading Finished')

    def get_columns_of_table(self, table_name):
        self.update_columns_dict()
        return self.columns_dict[table_name]

    def sql_query(self, query):

        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute(query)
        res = cur.fetchall()
        cur.close()
        conn.close()
        return res

    def sql_query_by_column_values(self, table_name, column, value_list):

        in_clause = ', '.join(value_list)
        res = self.sql_query(f"""select * from {table_name} where {column} IN ({in_clause})""")
        return res

    def create_index(self, table_name, columns):

        index_name = '__'.join(columns)

        sql = "CREATE INDEX IF NOT EXISTS %s ON %s(%s)" % (
            index_name,
            table_name,
            ', '.join(columns)
        )

        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute(sql)
        cur.close()
        conn.close()

    def close(self):
        conn = self.get_conn()
        conn.close()
